%% Implements Elixir quote.
-module(elixir_quote).
-export([escape/2, erl_escape/3, erl_quote/4,
         linify/2, linify/3, unquote/4, tail_join/3, join/2]).
-include("elixir.hrl").

%% Apply the line from site call on quoted contents.
linify(Line, Exprs) when is_integer(Line) ->
  do_linify(Line, nil, Exprs).

linify(Line, Var, Exprs) when is_integer(Line) ->
  do_linify(Line, Var, Exprs).

do_linify(Line, { Receiver, Counter } = Var, { Left, Meta, Receiver }) when is_atom(Left), is_list(Meta) ->
  NewMeta = case keyfind(counter, Meta) of
    { counter, _ } -> Meta;
    _ -> keystore(counter, Meta, Counter)
  end,
  do_tuple_linify(Line, Var, NewMeta, Left, Receiver);

do_linify(Line, Var, { Left, Meta, Right }) when is_list(Meta) ->
  do_tuple_linify(Line, Var, Meta, Left, Right);

do_linify(Line, Var, { Left, Right }) ->
  { do_linify(Line, Var, Left), do_linify(Line, Var, Right) };

do_linify(Line, Var, List) when is_list(List) ->
  [do_linify(Line, Var, X) || X <- List];

do_linify(_, _, Else) -> Else.

do_tuple_linify(Line, Var, Meta, Left, Right) ->
  NewMeta = case ?line(Meta) of
    0 -> keystore(line, Meta, Line);
    _ -> Meta
  end,
  { do_linify(Line, Var, Left), NewMeta, do_linify(Line, Var, Right) }.

%% Some expressions cannot be unquoted at compilation time.
%% This function is responsible for doing runtime unquoting.
unquote(Meta, Left, { '__aliases__', _, Args }, nil) ->
  { '__aliases__', Meta, [Left|Args] };

unquote(Meta, Left, Right, nil) when is_atom(Right) ->
  case atom_to_list(Right) of
    "Elixir." ++ _ ->
      { '__aliases__', Meta, [Left, Right] };
    _ ->
      { { '.', Meta, [Left, Right] }, Meta, [] }
  end;

unquote(Meta, Left, { Right, _, Context }, nil) when is_atom(Right), is_atom(Context) ->
  { { '.', Meta, [Left, Right] }, Meta, [] };

unquote(Meta, Left, { Right, _, Args }, nil) when is_atom(Right) ->
  { { '.', Meta, [Left, Right] }, Meta, Args };

unquote(_Meta, _Left, Right, nil) ->
  argument_error(<<"expected unquote after dot to return an atom, an alias or a quoted call, got: ",
                   ('Elixir.Macro':to_string(Right))/binary>>);

unquote(Meta, Left, Right, Args) when is_atom(Right) ->
  { { '.', Meta, [Left, Right] }, Meta, Args };

unquote(Meta, Left, { Right, _, Context }, Args) when is_atom(Right), is_atom(Context) ->
  { { '.', Meta, [Left, Right] }, Meta, Args };

unquote(_Meta, _Left, Right, _Args) ->
  argument_error(<<"expected unquote after dot with args to return an atom or a quoted call, got: ",
                   ('Elixir.Macro':to_string(Right))/binary>>).

join(Left, Right) when is_list(Right) ->
  validate_join(Left),
  Left ++ Right.

tail_join(Left, Right, Tail) when is_list(Right), is_list(Tail) ->
  validate_join(Left),
  Tail ++ Left ++ Right;

tail_join(Left, Right, Tail) when is_list(Left) ->
  validate_join(Left),
  [H|T] = lists:reverse(Tail ++ Left),
  lists:reverse([{ '|', [], [H, Right] }|T]).

validate_join(List) when is_list(List) ->
  ok;
validate_join(List) when not is_list(List) ->
  argument_error(<<"expected a list with quoted expressions in unquote_splicing/1, got: ",
                   ('Elixir.Kernel':inspect(List))/binary>>).

argument_error(Message) ->
  error('Elixir.ArgumentError':exception([{message,Message}])).

%% Escapes the given expression. It is similar to quote, but
%% lines are kept and hygiene mechanisms are disabled.
escape(Expr, Unquote) ->
  quote(Expr, nil, #elixir_quote{
    line=keep,
    vars_hygiene=false,
    aliases_hygiene=false,
    imports_hygiene=false,
    unquote=Unquote,
    escape=true
  }, nil).

erl_escape(Expr, Unquote, S) ->
  { QExpr, TQ } = escape(Expr, Unquote),
  { TExpr, TS } = elixir_translator:translate_each(QExpr, S),
  { TExpr, TQ, TS }.

%% Quotes an expression and return its quoted Elixir AST.

quote(Expr, nil, Q, S) ->
  do_quote(Expr, Q, S);

quote(Expr, Binding, Q, S) ->
  Context = Q#elixir_quote.context,

  Vars = [ { '{}', [],
    [ '=', [], [
      { '{}', [], [K, [], Context] },
      V
    ] ]
  } || { K, V } <- Binding],

  { TExprs, TQ } = do_quote(Expr, Q, S),
  { { '{}',[], ['__block__',[], Vars ++ [TExprs] ] }, TQ }.

erl_quote(Expr, Binding, Q, S) ->
  { QExpr, TQ } = quote(Expr, Binding, Q, S),
  { TExpr, TS } = elixir_translator:translate_each(QExpr, S),
  { TExpr, TQ, TS }.

%% Actual quoting and helpers

do_quote({ quote, _, Args } = Tuple, #elixir_quote{unquote=true} = Q, S) when length(Args) == 1; length(Args) == 2 ->
  { TTuple, TQ } = do_quote_tuple(Tuple, Q#elixir_quote{unquote=false}, S),
  { TTuple, TQ#elixir_quote{unquote=true} };

do_quote({ unquote, _Meta, [Expr] }, #elixir_quote{unquote=true} = Q, _) ->
  { Expr, Q#elixir_quote{unquoted=true} };

%% Context mark

do_quote({ defmodule, Meta, [_|_] = Args }, #elixir_quote{escape=false} = Q, S) ->
  %% Only store the context if we actually have a full alias as
  %% argument, otherwise the expression is being automatically
  %% generated and we don't want the alias to count.
  NewMeta =
    case hd(Args) of
      { '__aliases__', _, Atoms } ->
        case lists:all(fun is_atom/1, Atoms) of
          true  -> keystore(context, Meta, Q#elixir_quote.context);
          false -> Meta
        end;
      _ ->
        Meta
    end,

  do_quote_tuple({ defmodule, NewMeta, Args }, Q, S);

do_quote({ Def, Meta, [_|_] = Args }, #elixir_quote{escape=false} = Q, S) when ?defs(Def); Def == alias; Def == import ->
  NewMeta = keystore(context, Meta, Q#elixir_quote.context),
  do_quote_tuple({ Def, NewMeta, Args }, Q, S);

do_quote({ { '.', _, [_, Def] } = Target, Meta, Args }, #elixir_quote{escape=false} = Q, S) when ?defs(Def); Def == defmodule; Def == alias; Def == import ->
  NewMeta = keystore(context, Meta, Q#elixir_quote.context),
  do_quote_tuple({ Target, NewMeta, Args }, Q, S);

%% Aliases

do_quote({ 'alias!', _Meta, [Expr] }, #elixir_quote{aliases_hygiene=true} = Q, S) ->
  { TExpr, TQ } = do_quote(Expr, Q#elixir_quote{aliases_hygiene=false}, S),
  { TExpr, TQ#elixir_quote{aliases_hygiene=true} };

do_quote({ '__aliases__', Meta, [H|T] } = Alias, #elixir_quote{aliases_hygiene=true} = Q, S) when is_atom(H) and (H /= 'Elixir') ->
  Annotation = case elixir_aliases:expand(Alias, S#elixir_scope.aliases, S#elixir_scope.macro_aliases, S#elixir_scope.lexical_tracker) of
    Atom when is_atom(Atom) -> Atom;
    Aliases when is_list(Aliases) -> false
  end,
  AliasMeta = keystore(alias, Meta, Annotation),
  do_quote_tuple({ '__aliases__', AliasMeta, [H|T] }, Q, S);

%% Vars

do_quote({ Left, Meta, nil }, #elixir_quote{vars_hygiene=true} = Q, S) when is_atom(Left) ->
  do_quote_tuple({ Left, Meta, Q#elixir_quote.context }, Q, S);

%% Unquote

do_quote({ { { '.', Meta, [Left, unquote] }, _, [Expr] }, _, Args }, #elixir_quote{unquote=true} = Q, S) ->
  do_quote_call(Left, Meta, Expr, Args, Q, S);

do_quote({ { '.', Meta, [Left, unquote] }, _, [Expr] }, #elixir_quote{unquote=true} = Q, S) ->
  do_quote_call(Left, Meta, Expr, nil, Q, S);

%% Imports

do_quote({ '&', Meta, [{ '/', _, [{F, _, C}, A]}] = Args },
    #elixir_quote{imports_hygiene=true} = Q, S) when is_atom(F), is_integer(A), is_atom(C) ->
  do_quote_fa('&', Meta, Args, F, A, Q, S);

do_quote({ Name, Meta, ArgsOrAtom }, #elixir_quote{imports_hygiene=true} = Q, S) when is_atom(Name) ->
  Arity = case is_atom(ArgsOrAtom) of
    true  -> 0;
    false -> length(ArgsOrAtom)
  end,

  case (keyfind(import, Meta) == false) andalso
      elixir_dispatch:find_import(Meta, Name, Arity, S) of
    false ->
      AmbMeta =
        case (Arity == 1) andalso keyfind(ambiguous_op, Meta) of
          { ambiguous_op, nil } -> keystore(ambiguous_op, Meta, Q#elixir_quote.context);
          _ -> Meta
        end,
      do_quote_tuple({ Name, AmbMeta, ArgsOrAtom }, Q, S);
    Receiver ->
      ImportMeta = keystore(import,
        keystore(context, Meta, Q#elixir_quote.context),
        Receiver),
      do_quote_tuple({ Name, ImportMeta, ArgsOrAtom }, Q, S)
  end;

do_quote({ _, _, _ } = Tuple, #elixir_quote{escape=false} = Q, S) ->
  do_quote_tuple(Tuple, Q, S);

%% Literals

do_quote({ Left, Right }, #elixir_quote{unquote=true} = Q, S) when
    is_tuple(Left)  andalso (element(1, Left) == unquote_splicing);
    is_tuple(Right) andalso (element(1, Right) == unquote_splicing) ->
  do_quote({ '{}', [], [Left, Right] }, Q, S);

do_quote({ Left, Right }, Q, S) ->
  { TLeft, LQ }  = do_quote(Left, Q, S),
  { TRight, RQ } = do_quote(Right, LQ, S),
  { { TLeft, TRight }, RQ };

do_quote(Tuple, #elixir_quote{escape=true} = Q, S) when is_tuple(Tuple) ->
  { TT, TQ } = do_quote(tuple_to_list(Tuple), Q, S),
  { { '{}', [], TT }, TQ };

do_quote(List, #elixir_quote{escape=true} = Q, S) when is_list(List) ->
  % The improper case is pretty inefficient, but improper lists are hopefully
  % rare.
  case reverse_improper(List) of
    { L }       -> do_splice(L, Q, S);
    { L, R }    ->
      { TL, QL } = do_splice(L, Q, S, [], []),
      { TR, QR } = do_quote(R, QL, S),
      { update_last(TL, fun(X) -> { '|', [], [X, TR] } end), QR }
  end;
do_quote(List, Q, S) when is_list(List) ->
    do_splice(lists:reverse(List), Q, S);

do_quote(Other, Q, _) ->
  { Other, Q }.

%% Quote helpers

do_quote_call(Left, Meta, Expr, Args, Q, S) ->
  All  = [meta(Meta, Q), Left, { unquote, Meta, [Expr] }, Args],
  { TAll, TQ } = lists:mapfoldl(fun(X, Acc) -> do_quote(X, Acc, S) end, Q, All),
  { { { '.', Meta, [elixir_quote, unquote] }, Meta, TAll }, TQ }.

do_quote_fa(Target, Meta, Args, F, A, Q, S) ->
  NewMeta =
    case (keyfind(import_fa, Meta) == false) andalso
         elixir_dispatch:find_import(Meta, F, A, S) of
      false    -> Meta;
      Receiver -> keystore(import_fa, Meta, { Receiver, Q#elixir_quote.context })
    end,
  do_quote_tuple({ Target, NewMeta, Args }, Q, S).

do_quote_tuple({ Left, Meta, Right }, Q, S) ->
  { TLeft, LQ }  = do_quote(Left, Q, S),
  { TRight, RQ } = do_quote(Right, LQ, S),
  { { '{}', [], [TLeft, meta(Meta, Q), TRight] }, RQ }.

meta(Meta, #elixir_quote{line=keep}) ->
  Meta;
meta(Meta, #elixir_quote{line=nil}) ->
  keydelete(line, Meta);
meta(Meta, #elixir_quote{line=Line}) ->
  keystore(line, Meta, Line).

reverse_improper(L) -> reverse_improper(L, []).

reverse_improper([], Acc) -> { Acc };
reverse_improper([H|T], Acc) when is_list(T) -> reverse_improper(T, [H|Acc]);
reverse_improper([H|T], Acc) -> { [H|Acc], T }.

update_last([], _) -> [];
update_last([H], F) -> [F(H)];
update_last([H|T], F) -> [H|update_last(T,F)].

keyfind(Key, Meta) ->
  lists:keyfind(Key, 1, Meta).
keydelete(Key, Meta) ->
  lists:keydelete(Key, 1, Meta).
keystore(Key, Meta, Value) ->
  lists:keystore(Key, 1, Meta, { Key, Value }).

%% Quote splicing

do_splice([{ '|', Meta, [{ unquote_splicing, _, [Left] }, Right] }|T], #elixir_quote{unquote=true} = Q, S) ->
  %% Process the remaining entries on the list.
  %% For [1, 2, 3, unquote_splicing(arg)|tail], this will quote
  %% 1, 2 and 3, which could even be unquotes.
  { TT, QT } = do_splice(T, Q, S, [], []),
  { TR, QR } = do_quote(Right, QT, S),
  { do_runtime_join(Meta, tail_join, [Left, TR, TT]), QR#elixir_quote{unquoted=true} };

do_splice(List, Q, S) ->
  do_splice(List, Q, S, [], []).

do_splice([{ unquote_splicing, Meta, [Expr] }|T], #elixir_quote{unquote=true} = Q, S, Buffer, Acc) ->
  do_splice(T, Q#elixir_quote{unquoted=true}, S, [], do_runtime_join(Meta, join, [Expr, do_join(Buffer, Acc)]));

do_splice([H|T], Q, S, Buffer, Acc) ->
  { TH, TQ } = do_quote(H, Q, S),
  do_splice(T, TQ, S, [TH|Buffer], Acc);

do_splice([], Q, _S, Buffer, Acc) ->
  { do_join(Buffer, Acc), Q }.

do_join(Left, [])    -> Left;
do_join([], Right)   -> Right;
do_join(Left, Right) -> { { '.', [], ['Elixir.Kernel', '++'] }, [], [Left, Right] }.

do_runtime_join(Meta, Fun, Args) ->
  { { '.', Meta, [elixir_quote, Fun] }, Meta, Args }.
